#include <stdio.h>
#include <string.h>
#include <conio.h>

typedef unsigned char  uint8_t;
typedef unsigned short uint16_t;

uint16_t mpu_base = 0x330;
uint16_t sb_base = 0x220;

uint8_t dsp_ver_h;
uint8_t dsp_ver_l;

uint16_t error = 0;

#define SB_RESET_PORT   (sb_base + 0x6)
#define SB_READ_DATA    (sb_base + 0xA)
#define SB_WRITE_DATA   (sb_base + 0xC)
#define SB_WRITE_STATUS (sb_base + 0xC)
#define SB_READ_STATUS  (sb_base + 0xE)

#define ERROR_SB_RESET        (1)
#define ERROR_SB_SEND         (2)
#define ERROR_SB_RECV         (3)
#define ERROR_MPU_SEND_CMD    (4)
#define ERROR_MPU_SEND_DATA   (5)
#define ERROR_MPU_RECV_DATA   (6)
#define ERROR_MPU_LOOPBACK    (7)
#define ERROR_DSP_UNKNOWN     (8)
#define ERROR_DUMP_FAILED     (9)

char strbuf[256];
uint8_t memdump[256];

uint8_t dump[8192];

FILE *logfile;

void log_open (void) {
  logfile = fopen ("log.txt", "wt");
}

void log_close (void) {
  if (logfile != NULL) fclose (logfile);
}

void log (const char *str) {
  printf ("%s", str);
  if (logfile != NULL) fputs (str, logfile);
}



static uint8_t mpu_output_full (void) {
  return(inp(mpu_base + 1) & 0x40);
}
static unsigned char mpu_input_empty(void) {
  return(inp(mpu_base + 1) & 0x80);
}
static uint8_t mpu_read (void) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_input_empty() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_RECV_DATA; return 0; }
  return inp(mpu_base);
}
static void mpu_write (uint8_t byte) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_output_full() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_SEND_DATA; return; }
  outp (mpu_base, byte);
}
static void mpu_command (uint8_t cmd) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_output_full() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_SEND_CMD; return; }
  outp (mpu_base + 1, cmd);
}

void mpu_init (void) {
  mpu_command (0xFF);
  if (error) return;
  mpu_read();
  if (error) return;
  mpu_command (0x3F);
  if (error) return;
  mpu_read();
  if (error) return;
}

void mpu_loopback_check (void) {
  mpu_write (0xCA);
  if (error) return;
  if (mpu_read() != 0xCA) error = ERROR_MPU_LOOPBACK;
  if (error) return;
  mpu_write (0xFE);
  if (error) return;
  if (mpu_read() != 0xFE) error = ERROR_MPU_LOOPBACK;
  if (error) return;
}


static uint8_t sb_read_data_available (void) {
  return inp(SB_READ_STATUS) & 0x80;
}

static void sb_delay (uint16_t cycles) {  // 1 cycle is at least 480ns
  uint16_t i;
  for (i=0; i<cycles; ++i) inp(SB_WRITE_STATUS);
}

static void sb_send (uint8_t data) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if ((inp(SB_WRITE_STATUS) & 0x80) == 0x00) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_SEND; return; }

  outp (SB_WRITE_DATA, data);

}

static uint8_t sb_recv (void) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (inp(SB_READ_STATUS) & 0x80) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_RECV; return 0; }

  return inp (SB_READ_DATA);
}


void sb_reset (void) {
  uint16_t i;

  outp (SB_RESET_PORT, 1);
  sb_delay (7);  // more then 3us
  outp (SB_RESET_PORT, 0);
  for (i=0; i < 0xFFFF; ++i) { // 208 = ~100us
    if (sb_read_data_available()) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_RESET; return; }

  if (inp(SB_READ_DATA) != 0xAA) error = ERROR_SB_RESET;
}

void sb_get_dsp_ver (void) {
  uint8_t h,l;

  sb_send (0xE1);
  if (error) return;
  dsp_ver_h = sb_recv ();
  if (error) return;
  dsp_ver_l = sb_recv ();
  if (error) return;

}

static uint8_t sb_read_memory (uint8_t addr) {
  sb_send (0xF9);
  if (error) return 0;
  sb_send (addr);
  if (error) return 0;
  return sb_recv ();
}

static void sb_write_memory (uint8_t addr, uint8_t data) {
  sb_send (0xFA);
  if (error) return;
  sb_send (addr);
  if (error) return;
  sb_send (data);
}

void sb_clean_memory (uint8_t data) {
  int addr;

  for (addr = 0x40; addr < 256; ++addr) {
    sb_write_memory (addr, data);
    if (error) return;
  }
}

void sb_dump_memory (uint8_t *ptr) {
  int addr;

  for (addr = 0; addr < 256; ++addr) {
    ptr[addr] = sb_read_memory (addr);
    if (error) return;
  }
}

void sb_save_memdump (uint8_t *ptr, char *filename) {
  FILE *file;
  file = fopen (filename, "wb");
  if (file != NULL) {
    fwrite (ptr, 256, 1, file);
    fclose (file);
  }
}

void memory_dump (void) {
  sb_clean_memory (0xFF);
  if (error) return;
  sb_dump_memory (memdump);
  if (error) return;
  sb_save_memdump (memdump, "MDUMP2.bin");
  sb_clean_memory (0x00);
  if (error) return;
  sb_dump_memory (memdump);
  if (error) return;
  sb_save_memdump (memdump, "MDUMP1.bin");

}

void sb_initial_diag (void) {
  error = 0;
  log ("SB Reset: ");
  sb_reset ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  log ("DSP version: ");
  sb_get_dsp_ver ();
  if (error) { log ("failed!\n"); return; }
  else {
    sprintf (strbuf, "%d.%d\n", (int)dsp_ver_h, (int)dsp_ver_l);
    log (strbuf);
  }

  if (dsp_ver_h != 0x04) { error = ERROR_DSP_UNKNOWN; return; }


  log ("MPU-401 init: ");
  mpu_init ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  log ("Internal memory dump: ");
  memory_dump ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");


  log ("MPU-401 loopback check: ");
  mpu_loopback_check();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  switch (dsp_ver_l) {
    case 13:
    break;
    default:
    error = ERROR_DSP_UNKNOWN;
    return;
  }

}

void dsp_413_attack1 (uint16_t addr) {
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC4);

  sb_write_memory (0xBE, 0x04);
  sb_write_memory (0xBD, 0x18);

  sb_write_memory (0xBC, addr >> 8);
  sb_write_memory (0xBB, addr & 0xFF);

  sb_write_memory (0xB9, 0x11);
  sb_write_memory (0xB8, 0xE9);

  mpu_write (0x04);

  dump[addr] = sb_recv ();
  dump[addr + 1] = sb_recv ();
}

void dsp_413_attack2 (uint16_t addr) {
  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0xBE, addr >> 8);
  sb_write_memory (0xBD, addr & 0xFF);

  sb_write_memory (0xBB, 0x11);
  sb_write_memory (0xBA, 0xE9);

  sb_write_memory (0x01, 0xC2);

  mpu_write (0x00);

  sb_delay (67*2);

  sb_send (0xE1);

  dump[addr] = sb_recv ();
  dump[addr + 1] = sb_recv ();
}

void dsp_413_pentest (void) {
  error = 0;

  log ("Penetration test ... ");
  sb_reset ();
  mpu_init ();

  sb_write_memory (0xBF, 0x88);
  sb_write_memory (0xBC, 0xEE);


  sb_write_memory (0xBB, 0x0B);
  sb_write_memory (0xBA, 0x43);

  sb_write_memory (0x01, 0xC2);

  mpu_write (0x00);

  sb_delay (0xFFFF);

  sb_send (0xE1);

  sb_delay (0xFFFF);

  if (error == 0)   log ("prepared\n");
  else log ("failed to prepare\n");

  error = 0;
  sb_reset ();
  if (error == 0)   log ("reset done\n");
  else log ("reset failed\n");

  if (sb_read_memory (0x88) == 0xEE) log ("Penetration successful\n");
  else log ("Penetration failed\n");


}



void dsp_save_dump (void) {
  FILE *file;
  file = fopen ("dump.bin", "wb");
  if (file != NULL) {
    fwrite (dump, 8192, 1, file);
    fclose (file);
  }
}

/*
void dsp_413_dump (void) {
  
  uint16_t i;

  dump[0] = 0;
  dsp_413_attack (0x0000);
  if (error) { error = ERROR_DUMP_FAILED; return; }
  if (dump[0] != 0x02) { error = ERROR_DUMP_FAILED; return; }

  for (i=2; i<8192; i+=2) {
    dsp_413_attack (i);
    if (error) { error = ERROR_DUMP_FAILED; return; }
    if ((i & 0xFF) == 0x00) printf ("%04X dumped\n", i);
  }
  
  sb_reset ();
  log ("Dump complete !\n");
  dsp_save_dump ();

}
*/

void dsp_413_dump (void) {
  
  uint16_t i;

  log ("Attack started\n");
  dump[0] = 0;
  dsp_413_attack2 (0x0000);
  if (dump[0] != 0x02) {
    error = ERROR_DUMP_FAILED;
    return;
    }
  else log ("Attack successful\n");
  sb_reset ();
  dsp_save_dump ();
}


void error_print (void) {
  if (error) log ("ERROR: ");

  switch (error) {
    case ERROR_SB_RESET:
    log ("Cannot RESET sound blaster\n");
    break;
    case ERROR_SB_SEND:
    log ("Cannot SEND to sound blaster\n");
    break;
    case ERROR_SB_RECV:
    log ("Cannot RECV from sound blaster\n");
    break;
    case ERROR_DSP_UNKNOWN:
    log ("DSP version not supported\n");
    break;

    case ERROR_MPU_SEND_CMD :
    log ("Cannot SEND cmd to MPU\n");
    break;
    case ERROR_MPU_SEND_DATA:
    log ("Cannot SEND data to MPU\n");
    break;
    case ERROR_MPU_RECV_DATA:
    log ("Cannot RECV from MPU\n");
    break;
    case ERROR_MPU_LOOPBACK :
    log ("Midi loopback not present\n");
    break;

    case ERROR_DUMP_FAILED:
    log ("Dump failed\n");
    break;


    default:
    log ("Success!\n");

  }

}


void main (int argc, char *argv[]) {
  error = 0;
  log_open ();
  sb_initial_diag ();

  if (error == 0) {
    //dsp_413_dump ();
    dsp_413_pentest ();
  }

  sb_reset ();

  error_print ();
  log_close ();
}
